"""Converter classes for extensions to builtin types."""

import re

import yaml

from .standard import String, Integer, Float, Boolean
from .containers import Dictionary, List


# standard types with None as a default #######################################


class NoneString(String):

    """Converter for the `str` type with `None` as default."""

    DEFAULT = None


class NoneInteger(Integer):

    """Converter for the `int` type with `None` as default."""

    DEFAULT = None


class NoneFloat(Float):

    """Converter for the `float` type with `None` as default."""

    DEFAULT = None


class NoneBoolean(Boolean):

    """Converter for the `bool` type with `None` as default."""

    DEFAULT = None


# standard types with additional behavior #####################################


class _Literal(str):

    """Custom type for strings which should be dumped in the literal style."""

    @staticmethod
    def representer(dumper, data):
        """Return a custom dumper that formats `str` in the literal style."""
        return dumper.represent_scalar('tag:yaml.org,2002:str', data,
                                       style='|' if data else '')

yaml.add_representer(_Literal, _Literal.representer)


class Markdown(String):

    """Converter for a `str` type that contains Markdown."""

    REGEX_MARKDOWN_SPACES = re.compile(r"""

    ([^\n ])  # any character but a newline or space

    (\ ?\n)   # optional space + single newline

    (?!       # none of the following:

      (?:\s)       # whitespace
      |
      (?:[-+*]\s)  # unordered list separator + whitespace
      |
      (?:\d+\.\s)  # number + period + whitespace

    )

    ([^\n])  # any character but a newline

    """, re.VERBOSE | re.IGNORECASE)

    # based on: http://en.wikipedia.org/wiki/Sentence_boundary_disambiguation
    REGEX_SENTENCE_BOUNDARIES = re.compile(r"""

    (            # one of the following:

      (?<=[a-z)][.?!])      # lowercase letter + punctuation
      |
      (?<=[a-z0-9][.?!]\")  # lowercase letter/number + punctuation + quote

    )

    (\s)          # any whitespace

    (?=\"?[A-Z])  # optional quote + an upppercase letter

    """, re.VERBOSE)

    @classmethod
    def to_value(cls, obj):
        """Join non-meaningful line breaks."""
        value = String.to_value(obj)
        return Markdown.join(value)

    @staticmethod
    def join(text):
        r"""Convert single newlines (ignored by Markdown) to spaces.

        >>> Markdown.join("abc\n123")
        'abc 123'

        >>> Markdown.join("abc\n\n123")
        'abc\n\n123'

        >>> Markdown.join("abc \n123")
        'abc 123'

        """
        return Markdown.REGEX_MARKDOWN_SPACES.sub(r'\1 \3', text).strip()

    @classmethod
    def to_data(cls, obj):
        """Break a string at sentences and dump as a literal string."""
        value = String.to_value(obj)
        data = String.to_data(value)
        split = Markdown.split(data)
        return _Literal(split)

    @staticmethod
    def split(text, end='\n'):
        r"""Replace sentence boundaries with newlines and append a newline.

        :param text: string to line break at sentences
        :param end: appended to the end of the update text

        >>> Markdown.split("Hello, world!", end='')
        'Hello, world!'

        >>> Markdown.split("Hello, world! How are you? I'm fine. Good.")
        "Hello, world!\nHow are you?\nI'm fine.\nGood.\n"

        """
        stripped = text.strip()
        if stripped:
            return Markdown.REGEX_SENTENCE_BOUNDARIES.sub('\n', stripped) + end
        else:
            return ''

# container types with additional behavior ####################################


class AttributeDictionary(Dictionary):

    """Dictionary converter with keys available as attributes."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__dict__ = self

    @classmethod
    def create_default(cls):
        """Create an uninitialized object with keys as attributes."""
        if cls is AttributeDictionary:
            msg = "AttributeDictionary class must be subclassed to use"
            raise NotImplementedError(msg)

        obj = cls.__new__(cls)
        obj.__dict__ = obj
        return obj


class SortedList(List):

    """List converter that is sorted on disk."""

    @classmethod
    def create_default(cls):
        """Create an uninitialized object."""
        if cls is SortedList:
            msg = "SortedList class must be subclassed to use"
            raise NotImplementedError(msg)
        if not cls.item_type:
            msg = "SortedList subclass must specify item type"
            raise NotImplementedError(msg)

        return cls.__new__(cls)

    @classmethod
    def to_data(cls, obj):
        """Convert all attribute values for optimal dumping to YAML."""
        value = cls.to_value(obj)

        data = []

        for item in sorted(value):
            data.append(cls.item_type.to_data(item))

        return data



#!/usr/bin/env python
# pylint:disable=R0201,C0111

"""Integration tests for the `yorm.settings.fake` option."""

import os
import pytest
from unittest.mock import patch

import yorm

from . import strip


@yorm.attr(value=yorm.converters.standard.Integer)
@yorm.sync("path/to/{self.name}.yml")
class Sample:

    """Sample class for fake mapping."""

    def __init__(self, name):
        self.name = name
        self.value = 0

    def __repr__(self):
        return "<sample {}>".format(id(self))


@patch('yorm.settings.fake', True)
class TestFake:

    """Integration tests with `yorm.settings.fake` enabled."""

    def test_no_file_create_when_fake(self, tmpdir):
        tmpdir.chdir()
        sample = Sample('sample')

        # ensure no file is created
        assert "path/to/sample.yml" == sample.yorm_mapper.path
        assert not os.path.exists(sample.yorm_mapper.path)

        # change object values
        sample.value = 42

        # check fake file
        assert strip("""
        value: 42
        """) == sample.yorm_mapper.text

        # ensure no file is created
        assert not os.path.exists(sample.yorm_mapper.path)

        # change fake file
        sample.yorm_mapper.text = "value: 0\n"

        # check object values
        assert 0 == sample.value

        # ensure no file is created
        assert not os.path.exists(sample.yorm_mapper.path)

    def test_fake_changes_indicate_modified(self, tmpdir):
        tmpdir.chdir()
        sample = Sample('sample')

        assert False is sample.yorm_mapper.modified
        assert 0 == sample.value

        sample.yorm_mapper.text = "value: 42\n"

        assert True is sample.yorm_mapper.modified
        assert 42 == sample.value
        assert False is sample.yorm_mapper.modified


class TestReal:

    """Integration tests with `yorm.settings.fake` disabled."""

    def test_setting_text_updates_attributes(self, tmpdir):
        tmpdir.chdir()
        sample = Sample('sample')

        sample.yorm_mapper.text = "value: 42"

        assert 42 == sample.value

    def test_setting_attributes_update_text(self, tmpdir):
        tmpdir.chdir()
        sample = Sample('sample')

        sample.value = 42

        assert strip("""
        value: 42
        """) == sample.yorm_mapper.text


if __name__ == '__main__':
    pytest.main()



#!/usr/bin/env python
# pylint: disable=R,C

"""Unit tests for the `mapper` module."""

import os
import pytest
from unittest.mock import patch, Mock

from yorm.mapper import Helper, Mapper
from yorm.converters import Integer


@patch('yorm.settings.fake', True)
class TestHelperFake:

    """Unit tests for fake mappings using the `Helper` class."""

    def test_create(self):
        """Verify fake files can be created."""
        mapper = Helper("fake/path/to/file")
        mapper.create(Mock())

        assert not os.path.exists(mapper.path)

    def test_delete(self):
        """Verify fake files can be deleted."""
        mapper = Helper("fake/path/to/file")
        mapper.create(None)
        mapper.delete()

        assert not os.path.exists(mapper.path)

    def test_modified(self):
        """Verify fake files can be modified."""
        mapper = Helper("fake/path/to/file")
        assert mapper.modified

        mapper.create(None)
        assert mapper.modified

        mapper.modified = False
        assert not mapper.modified

        mapper.modified = True
        assert mapper.modified


class TestHelperReal:

    """Unit tests for real mappings using the `Helper` class."""

    def test_create(self, tmpdir):
        """Verify files can be created."""
        tmpdir.chdir()
        mapper = Helper("real/path/to/file")
        mapper.create(None)

        assert os.path.isfile(mapper.path)

    def test_create_twice(self, tmpdir):
        """Verify the second creation is ignored."""
        tmpdir.chdir()
        mapper = Helper("real/path/to/file")
        mapper.create(None)
        mapper.create(None)

        assert os.path.isfile(mapper.path)

    def test_delete(self, tmpdir):
        """Verify files can be deleted."""
        tmpdir.chdir()
        mapper = Helper("real/path/to/file")
        mapper.create(None)
        mapper.delete()

        assert not os.path.exists(mapper.path)

    def test_delete_twice(self, tmpdir):
        """Verify the second deletion is ignored."""
        tmpdir.chdir()
        mapper = Helper("real/path/to/file")
        mapper.delete()

        assert not os.path.exists(mapper.path)

    def test_modified(self, tmpdir):
        """Verify files track modifications."""
        tmpdir.chdir()
        mapper = Helper("real/path/to/file")
        assert mapper.modified

        mapper.create(None)
        assert mapper.modified

        mapper.modified = False
        assert not mapper.modified

        mapper.modified = True
        assert mapper.modified

    def test_modified_deleted(self):
        """Verify a deleted file is always modified."""
        mapper = Helper("fake/path/to/file")

        assert mapper.modified


class TestMapper:

    """Unit tests for the `Mapper` class."""

    class MyObject:
        foo = 1

    def test_auto_off(self, tmpdir):
        """Verify storage is delayed with auto off."""
        tmpdir.chdir()
        obj = self.MyObject()
        attrs = {'number': Integer}
        mapper = Mapper(obj, "real/path/to/file", attrs, auto=False)
        assert False is mapper.auto

        mapper.create()
        assert "" == mapper.text
        assert False is mapper.auto

        mapper.store()
        assert "" == mapper.text
        assert False is mapper.auto

        mapper.store(force=True)
        assert "number: 0\n" == mapper.text
        assert False is mapper.auto

    def test_missing_attributes_added(self):
        obj = self.MyObject()
        path = "mock/path"
        attrs = {'bar': Integer, 'qux': Integer}
        mapper = Mapper(obj, path, attrs)
        mapper.create()
        mapper.fetch()

        assert 1 == obj.foo
        assert 0 == obj.bar
        assert 0 == obj.qux


if __name__ == '__main__':
    pytest.main()
